﻿using System;
namespace PID_Test
{
    public class PID
    {
        public enum PidModeType
        {
            PID_Mode_Automatic = 1,
            PID_Mode_Manual = 0
        };

        public enum PidDirectionType
        {
            PID_Direction_Direct = 0,
            PID_Direction_Reverse = 1
        };

        public struct PidType
        {
            public double dispKp; // * we'll hold on to the tuning parameters in user-entered
            public double dispKi; //   format for display purposes
            public double dispKd; //

            public double kp; // * (P)roportional Tuning Parameter
            public double ki; // * (I)ntegral Tuning Parameter
            public double kd; // * (D)erivative Tuning Parameter

            public PidDirectionType controllerDirection;

            public double myInput; // * Pointers to the Input, Output, and Setpoint variables
            public double myOutput; //   This creates a hard link between the variables and the
            public double mySetpoint; //   PID, freeing the user from having to constantly tell us
            //   what these values are.  with pointers we'll just know.

            public double ITerm, lastInput;

            public UInt32 SampleTime;
            public double outMin, outMax;
            public bool inAuto;
        };

        public PidType pid;

        public void PID_init(double Kp, double Ki, double Kd, PidDirectionType ControllerDirection)
        {
            pid.myInput = 0;
            pid.myOutput = 0;
            pid.mySetpoint = 0;
            pid.ITerm = 0;
            pid.lastInput = 0;
            pid.inAuto = false;

            PID_SetOutputLimits(0, (double)0xFFFF);

            //default Controller Sample Time is 1 seconds
            pid.SampleTime = 1000;

            PID_SetControllerDirection(ControllerDirection);
            PID_SetTunings(Kp, Ki, Kd);
        }

        public bool PID_Compute()
        {
            if (!pid.inAuto) return false;

            double input = pid.myInput;
            double error = pid.mySetpoint - input;
            pid.ITerm += (pid.ki * error);
            if (pid.ITerm > pid.outMax)
                pid.ITerm = pid.outMax;
            else if (pid.ITerm < pid.outMin)
                pid.ITerm = pid.outMin;
            double dInput = (input - pid.lastInput);

            /*Compute PID Output*/
            double output = pid.kp * error + pid.ITerm - pid.kd * dInput;

            if (output > pid.outMax)
                output = pid.outMax;
            else if (output < pid.outMin)
                output = pid.outMin;
            pid.myOutput = output;

            /*Remember some variables for next time*/
            pid.lastInput = input;
            return true;
        }

        void PID_SetTunings(double Kp, double Ki, double Kd)
        {
            if (Kp < 0 || Ki < 0 || Kd < 0) return;

            pid.dispKp = Kp;
            pid.dispKi = Ki;
            pid.dispKd = Kd;

            double SampleTimeInSec = ((double)pid.SampleTime) / 1000;
            pid.kp = Kp;
            pid.ki = Ki * SampleTimeInSec;
            pid.kd = Kd / SampleTimeInSec;

            if (pid.controllerDirection == PidDirectionType.PID_Direction_Reverse)
            {
                pid.kp = (0 - pid.kp);
                pid.ki = (0 - pid.ki);
                pid.kd = (0 - pid.kd);
            }
        }

        void PID_SetSampleTime(int NewSampleTime) 
        {
            if (NewSampleTime > 0) 
            {
                double ratio = (double) NewSampleTime / (double) pid.SampleTime;
                pid.ki *= ratio;
                pid.kd /= ratio;
                pid.SampleTime = (UInt32) NewSampleTime;
            }
        }

        public void PID_SetOutputLimits(double Min, double Max)
        {
            if (Min >= Max) return;          
            pid.outMin = Min;
            pid.outMax = Max;

            if (pid.inAuto)
            {
                if (pid.myOutput > pid.outMax)
                {
                    pid.myOutput = pid.outMax;
                }
                else if (pid.myOutput < pid.outMin)
                {
                    pid.myOutput = pid.outMin;
                }

                if (pid.ITerm > pid.outMax)
                {
                    pid.ITerm = pid.outMax;
                }
                else if (pid.ITerm < pid.outMin)
                {
                    pid.ITerm = pid.outMin;
                }
            }
        }

        public void PID_SetMode(PidModeType Mode)
        {
            bool newAuto = (Mode == PidModeType.PID_Mode_Automatic);
            if (newAuto != pid.inAuto)
            {  /*we just went from manual to auto*/
                PID_Initialize();
            }
            pid.inAuto = newAuto;
        }

        void PID_Initialize()
        {
            pid.ITerm = pid.myOutput;
            pid.lastInput = pid.myInput;
            if (pid.ITerm > pid.outMax)
            {
                pid.ITerm = pid.outMax;
            }
            else if (pid.ITerm < pid.outMin)
            {
                pid.ITerm = pid.outMin;
            }
        }

        void PID_SetControllerDirection(PidDirectionType Direction)
        {
            if (pid.inAuto && Direction != pid.controllerDirection)
            {
                pid.kp = (0 - pid.kp);
                pid.ki = (0 - pid.ki);
                pid.kd = (0 - pid.kd);
            }
            pid.controllerDirection = Direction;
        }
    }
}
